# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import unicode_literals

import six
from .DocumentPath import DocumentPath

class ItemBuilder(object):
    def __init__(self):
        self._parts = []

    def build(self):
        # This ensures that paths with different list indexes are reconstructed
        # in document order, not request order
        self._parts.sort()

        # It's marginally simpler to start with a fake map and throw the top
        # level away than handle the top-level case
        item = {'M': {}}
        for path, av in self._parts:
            _merge(item, _build_path(path, av), 'M')

        _fixup_lists(item)

        return item['M']

    def with_value(self, path, av):
        if isinstance(path, DocumentPath):
            path = path.elements

        self._parts.append((path, av))

def _merge(existing, new, t):
    for k in new[t]:
        if k not in existing[t]:
            existing[t][k] = new[t][k]
        else:
            if _is_map(existing[t][k]):
                _merge(existing[t][k], new[t][k], 'M')
            elif _is_list(existing[t][k]):
                _merge(existing[t][k], new[t][k], 'L')
            else:
                raise TypeError("Unexpected type: " + type(existing[t][k]).__name__)

def _build_path(path, av):
    if len(path) == 0:
        return av
    else:
        head, tail = path[0], path[1:]
        e = _build_value(head, _build_path(tail, av))
        return e

def _build_value(index, value):
    if _is_key(index):
        return {'M': {index: value}}
    elif _is_ordinal(index):
        # Store the index in case any other paths are the same index into the list
        return {'L': {index: value}}
    else:
        raise TypeError('index must be a string or integer type, got {}'.format(type(index).__name__))

def _fixup_lists(item):
    if _is_list(item):
        item['L'] = _flatten_list(item['L'])
    elif _is_map(item):
        for value in item['M'].values():
            _fixup_lists(value)
    else:
        # Leaf AV, stop
        return

def _flatten_list(ml):

    if isinstance(ml, dict):
        # Flatten a map list into a list sorted by index
        flat = []
        for _, value in sorted(ml.items()):
            _fixup_lists(value)
            flat.append(value)
        return flat
    else:
        # Nevermind, it's a regular list, just return it
        return ml

def _is_key(index):
    return isinstance(index, six.string_types)

def _is_ordinal(index):
    return isinstance(index, six.integer_types)

def _is_map(c):
    return len(c) == 1 and 'M' in c

def _is_list(c):
    return len(c) == 1 and 'L' in c



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import threading

from .Tube import SocketTubePool
from .DaxClient import DaxClient
from .DaxError import DaxClientError, DaxErrorCode
from .Router import EndpointRouter

import logging
logger = logging.getLogger(__name__)

DEFAULT_PORT = 8111

class Cluster(object):
    def __init__(self, region_name, discovery_endpoints, credentials, user_agent=None, user_agent_extra=None, connect_timeout=None,
                 read_timeout=None, router_factory=None, client_factory=None):
        self._region_name = region_name
        self._discovery_endpoints = [_parse_host_ports(endpoint) for endpoint in discovery_endpoints]
        self._credentials = credentials
        self._user_agent = user_agent
        self._user_agent_extra = user_agent_extra
        self._connect_timeout = connect_timeout
        self._read_timeout = read_timeout
        self._client_factory = client_factory or self._new_client

        _router_factory = router_factory or EndpointRouter

        self._closed = False

        # All time intervals are in seconds to match time.time
        self._cluster_update_interval = 4.0
        self._health_check_interval = 5.0
        self._idle_connection_reap_delay = 30.0

        self._router = _router_factory(self._client_factory,
                                       self._discovery_endpoints,
                                       self._cluster_update_interval,
                                       self._health_check_interval)

        self._lock = threading.Lock()

    def start(self, min_healthy=1):
        self._router.start()
        self.wait_for_routes(min_healthy=min_healthy, leader_min=1, timeout=self._connect_timeout)

    def close(self):
        if self._closed:
            return

        self._closed = True

        try:
            self._router.close()
        except Exception as e: # pylint: disable=broad-except
            logger.warning('Failed closing router', exec_info=e)
        finally:
            self._router = None

    def read_client(self, prev_client=None):
        ''' Return a read client.

        Caller should not close the client.
        '''
        client = self._router.next_any(prev_client)
        if client is None:
            raise DaxClientError("No cluster endpoints available", DaxErrorCode.NoRoute)

        return client

    def write_client(self, prev_client=None):
        ''' Return a write client.

        Caller should not close the client.
        '''
        client = self._router.next_leader(prev_client)
        if client is None:
            raise DaxClientError("No cluster endpoints available", DaxErrorCode.NoRoute)

        return client

    def wait_for_routes(self, min_healthy=1, leader_min=1, timeout=None):
        return self._router.wait_for_routes(min_healthy=min_healthy, leader_min=leader_min, timeout=timeout)

    def _new_client(self, host, port):
        ''' Create a new client.

        Caller is responsible for closing the client.
        '''
        tube_pool = SocketTubePool(
                host, port,
                lambda: self._credentials,
                self._region_name,
                self._user_agent,
                self._user_agent_extra,
                self._connect_timeout,
                self._read_timeout)
        return DaxClient(tube_pool)

def _parse_host_ports(endpoint):
    parts = endpoint.split(':', 1)
    if len(parts) == 1:
        return parts[0].strip(), DEFAULT_PORT
    else:
        return parts[0].strip(), int(parts[1].strip())



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .DaxError import DaxClientError, DaxErrorCode

class DocumentPath(object):
    def __init__(self, elements):
        self.elements = elements

    def __eq__(self, other):
        return self.elements == other.elements

    def __repr__(self):
        return 'DocumentPath(' + repr(self.elements) + ')'

    @classmethod
    def from_path(cls, path, attr_names=None):
        attr_names = attr_names or {}

        split = path.split('.')
        elements = []

        # TODO Rather than mutating element, would it be faster to 
        for element in split:
            index = element.find('[')
            if index == -1:
                elements.append(attr_names.get(element, element))
                continue
            elif index == 0:
                raise DaxClientError('Invalid path: ' + path, DaxErrorCode.Validation, False)

            initial = element[:index]
            elements.append(attr_names.get(initial, initial))

            while index != -1:
                element = element[index+1:]
                index = element.find(']')

                if index == -1:
                    raise DaxClientError('Invalid path: ' + path, DaxErrorCode.Validation, False)

                ordinal = int(element[:index])
                elements.append(ordinal)

                element = element[index+1:]
                index = element.find('[')

                if index > 0:
                    raise DaxClientError('Invalid path: ' + path, DaxErrorCode.Validation, False)

            if element:
                raise DaxClientError('Invalid path: ' + path, DaxErrorCode.Validation, False)

        return cls(elements)




# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging

from .Cluster import Cluster
from .RequestValidator import RequestValidator
from .RetryHandler import RetryHandler

import botocore.session
from botocore.client import Config, ClientMeta
from botocore.credentials import Credentials
from botocore.model import ServiceModel
from botocore.exceptions import PartialCredentialsError
from botocore.hooks import first_non_none_response
from .Constants import PY_TO_OP_NAME

logger = logging.getLogger(__name__)


class AmazonDaxClient(object):
    ''' Construct a DAX Client.

    Use
        session = botocore.session.get_session()
        dax = AmazonDaxClient(session, ...)

    instead of
        session = botocore.session.get_session()
        ddb = session.create_client('dynamodb', ...)

    Remember to close the cluster when done, or use a with-statement:
        with AmazonDaxClient(session, ...) as dax:
            ...

    To pass the endpoints, use the 'endpoint_url' or 'endpoints' parameters:

        AmazonDaxClient(endpoint_url='myDAXcluster.2cmrwl.clustercfg.dax.use1.cache.amazonaws.com:8111')
        AmazonDaxClient(endpoints=['myDAXcluster.2cmrwl.clustercfg.dax.use1.cache.amazonaws.com:8111'])

    AmazonDaxClient instances can be shared between threads.
    '''

    @staticmethod
    def resource(session=None,
                 region_name=None, api_version=None,
                 use_ssl=False, verify=None, endpoint_url=None,
                 aws_access_key_id=None, aws_secret_access_key=None,
                 aws_session_token=None, config=None, endpoints=None):
        ''' Create a high-level Boto Resource matching DynamoDB.

        Instead of
            ddb = boto3.resource('dynamodb')

        use
            dax = AmazonDaxClient.resource()
        '''
        # Local import so it doesn't fail if boto3 is not available
        from .Resource import is_boto3_session, DaxSession

        if session is None or isinstance(session, botocore.session.Session):
            session = DaxSession(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name,
                botocore_session=session,
                endpoints=endpoints)
        else:
            if is_boto3_session(session):
                session = DaxSession.from_boto3(session, endpoints)
            else:
                raise ValueError('session must be a botocore or boto3 session')

        # Create the resource
        res = session.resource('dynamodb', region_name=region_name, api_version=api_version,
                               use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
                               aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key,
                               aws_session_token=aws_session_token, config=config)

        return res

    def __init__(self, session=None,
                 region_name=None, api_version=None,
                 use_ssl=None, verify=None, endpoint_url=None,
                 aws_access_key_id=None, aws_secret_access_key=None,
                 aws_session_token=None, config=None, endpoints=None):
        if session is None:
            session = botocore.session.get_session()
        else:
            if not isinstance(session, botocore.session.Session):
                try:
                    # Check for a boto3 session-ish object and get the internal botocore session
                    _session = session._session
                except AttributeError:
                    raise ValueError(
                        'session must be a botocore or boto3 session')
                else:
                    session = _session

        if use_ssl:
            raise ValueError('SSL/TLS is not supported. Set use_ssl=False.')

        self._session = session
        default_client_config = self._session.get_default_client_config()
        if config is not None and default_client_config is not None:
            # If a config is provided and a default config is set, then
            # use the config resulting from merging the two.
            config = default_client_config.merge(config)
        elif default_client_config is not None:
            # If a config was not provided then use the default
            # client config from the session
            config = default_client_config
        self._client_config = config if config is not None else Config(
            region_name=region_name)

        # resolve the region name
        if region_name is None:
            if config and config.region_name:
                region_name = config.region_name
            else:
                region_name = self._session.get_config_variable('region')
        self._region_name = region_name

        # Figure out the verify value base on the various
        # configuration options.
        if verify is None:
            verify = self._session.get_config_variable('ca_bundle')
        self._verify = verify

        # Gather endpoints
        self._endpoints = endpoints or []
        if endpoint_url and endpoint_url not in self._endpoints:
            # If endpoint_url is provided, include it
            self._endpoints.insert(0, endpoint_url)

        if not self._endpoints:
            raise ValueError('No endpoints provided')

        # Resolve credentials
        if aws_access_key_id is not None and aws_secret_access_key is not None:
            self._credentials = Credentials(
                aws_access_key_id, aws_secret_access_key, aws_session_token)
        elif self._session._missing_cred_vars(aws_access_key_id, aws_secret_access_key):
            raise PartialCredentialsError(
                provider='explicit',
                cred_var=self._session._missing_cred_vars(aws_access_key_id, aws_secret_access_key))
        else:
            self._credentials = self._session.get_credentials()

        # Fake out the meta information as much as possible
        loader = session.get_component('data_loader')
        json_model = loader.load_service_model(
            'dynamodb', 'service-2', api_version=api_version)
        service_model = ServiceModel(json_model, service_name='dynamodb')
        event_emitter = session.get_component('event_emitter')
        partition = None
        self.meta = ClientMeta(event_emitter, self._client_config,
                               self._endpoints[0], service_model, PY_TO_OP_NAME, partition)

        # Check signing version
        if self._client_config.signature_version and self._client_config.signature_version != 'v4':
            logger.warning('DAX only supports SigV4 signing; given signature_version "%s" ignored.',
                           self._client_config.signature_version)

        # Start cluster connection & background tasks
        self._cluster = Cluster(self._region_name,
                                self._endpoints,
                                self._credentials,
                                self._client_config.user_agent,
                                self._client_config.user_agent_extra,
                                self._client_config.connect_timeout,
                                self._client_config.read_timeout)
        self._cluster.start()

    def close(self):
        if self._cluster:
            self._cluster.close()
            self._cluster = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def batch_get_item(self, **kwargs):
        return self._read_request('batch_get_item', **kwargs)

    def batch_write_item(self, **kwargs):
        return self._write_request('batch_write_item', **kwargs)

    def delete_item(self, **kwargs):
        return self._write_request('delete_item', **kwargs)

    def get_item(self, **kwargs):
        return self._read_request('get_item', **kwargs)

    def put_item(self, **kwargs):
        return self._write_request('put_item', **kwargs)

    def query(self, **kwargs):
        return self._read_request('query', **kwargs)

    def scan(self, **kwargs):
        return self._read_request('scan', **kwargs)

    def update_item(self, **kwargs):
        return self._write_request('update_item', **kwargs)

    def transact_get_items(self, **kwargs):
        return self._read_request('transact_get_items', **kwargs)

    def transact_write_items(self, **kwargs):
        return self._write_request('transact_write_items', **kwargs)

    # vvv TODO Boto service methods vvv
    def can_paginate(self, **kwargs):
        raise NotImplementedError()

    def generate_presigned_url(self, **kwargs):
        raise NotImplementedError()

    def get_paginator(self, **kwargs):
        raise NotImplementedError()

    def get_waiter(self, **kwargs):
        raise NotImplementedError()

    # vvv Internal Methods vvv

    def _read_request(self, op_name, **kwargs):
        rclient = self._cluster.read_client()
        return self._retryable_request(rclient, op_name, **kwargs)

    def _write_request(self, op_name, **kwargs):
        wclient = self._cluster.write_client()
        return self._retryable_request(wclient, op_name, **kwargs)

    def _retryable_request(self, client, op_name, **kwargs):
        action = getattr(client, op_name)
        retryer = RetryHandler(self._client_config, self._cluster)
        while True:
            response = None
            try:
                response = self._do_request(action, op_name, **kwargs)
            except Exception as e: # pylint: disable=broad-except
                retryer.on_exception(e)
                if retryer.can_retry(e):
                    retryer.pause_before_retry(e)
                    continue
                else:
                    raise e

            return response

    def _do_request(self, action, op_name, **kwargs):
        api_params = kwargs
        operation_model = self.meta.service_model.operation_model(
            self.meta.method_to_api_mapping[op_name])
        request_context = {
            'client_region': self.meta.region_name,
            'client_config': self.meta.config,
            'has_streaming_input': operation_model.has_streaming_input,
            'auth_type': operation_model.auth_type,
        }

        # Copy boto3's events so that the resource works properly

        # Emit an event that allows users to modify the parameters at the
        # beginning of the method. It allows handlers to modify existing
        # parameters or return a new set of parameters to use.
        responses = self.meta.events.emit(
            'provide-client-params.dynamodb.{operation_name}'.format(
                operation_name=operation_model.name),
            params=api_params, model=operation_model, context=request_context)
        api_params = first_non_none_response(responses, default=api_params)
        self.meta.events.emit(
            'before-parameter-build.dynamodb.{operation_name}'.format(
                operation_name=operation_model.name),
            params=api_params, model=operation_model, context=request_context)
        RequestValidator.validate_api_using_operation_model(
            operation_model, api_params)

        response = action(**api_params)

        # Ignore the HTTP response and just handle the parsed response
        self.meta.events.emit(
            'after-call.dynamodb.{operation_name}'.format(
                operation_name=operation_model.name),
            http_response=None, parsed=response,
            model=operation_model, context=request_context
        )

        return response

    # vvv Unsupported Table/Service Methods vvv
    def create_backup(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def create_global_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def create_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def delete_backup(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def delete_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def describe_backup(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def describe_continuous_backups(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def describe_global_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def describe_limits(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def describe_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def describe_time_to_live(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def list_backups(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def list_global_tables(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def list_tables(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def list_tags_of_resource(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def restore_table_from_backup(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def tag_resource(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def untag_resource(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def update_global_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def update_table(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")

    def update_time_to_live(self, **kwargs):
        raise NotImplementedError(
            "Table operations are not supported on DAX. Use the boto DynamoDB client instead.")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from . import AttributeValueEncoder, AttributeValueDecoder, Assemblers, CborTypes, DynamoDbV1Converter
from .Cache import SimpleCache, RefreshingCache
from . import CborDecoder
from .generated import Stubs
from .DaxError import DaxServiceError
from .Constants import DaxMethodIds, DynamoDBOperationNumerals, ReturnValuesOnConditionCheckFailure
from .RequestValidator import RequestValidator
from copy import deepcopy

import logging

logger = logging.getLogger(__name__)


class Request(dict):
    def __init__(self, d=None, **kwargs):
        super(Request, self).__init__()
        if d:
            self.update(d)
        if kwargs:
            self.update(kwargs)
        self.attr_names = None
        self.attr_list_id = None
        self.key_schema = None
        self.key_schema_by_table = None

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value


class DaxClient(object):
    SUCCESS = CborTypes.TYPE_ARRAY + 0  # Success is an empty array

    CACHE_SIZE = 1000
    KEY_CACHE_TTL_MILLIS = 60000

    def __init__(self, tube_pool):
        self._tube_pool = tube_pool

        # Caches are per-node for simplicity
        self._key_cache = RefreshingCache(DaxClient.CACHE_SIZE, self._defineKeySchema, DaxClient.KEY_CACHE_TTL_MILLIS)
        self._attr_list_cache = SimpleCache(DaxClient.CACHE_SIZE, self._defineAttributeList)
        self._attr_list_id_cache = SimpleCache(DaxClient.CACHE_SIZE, self._defineAttributeListId)

    def close(self):
        if self._tube_pool:
            self._tube_pool.close()
            self._tube_pool = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def batch_get_item(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.batchGetItem_N697851100_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema_by_table = {tableName: self._key_cache.get(tableName, tube) \
                                            for tableName in request.RequestItems}

            logger.debug('Sending BatchGetItem request')
            Stubs.write_batchGetItem_N697851100_1(request, tube)

            result = self._decode_result('BatchGetItem', request, Assemblers.batchGetItem_N697851100_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def batch_write_item(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.batchWriteItem_116217951_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema_by_table = {tableName: self._key_cache.get(tableName, tube) \
                                            for tableName in request.RequestItems}

            # Attach AttrListIDs to each PutRequest for later encoding
            for table_name, write_requests in request.RequestItems.items():
                for write_request in write_requests:
                    for write_type, write_info in write_request.items():
                        if write_type == 'PutRequest':
                            attr_names = AttributeValueEncoder._get_canonical_attribute_list(
                                write_info['Item'],
                                request.key_schema_by_table[table_name])
                            write_info['attr_names'] = attr_names
                            write_info['attr_list_id'] = self._attr_list_id_cache.get(attr_names, tube)

            logger.debug('Sending BatchWriteItem request')
            Stubs.write_batchWriteItem_116217951_1(request, tube)

            result = self._decode_result('BatchWriteItem', request, Assemblers.batchWriteItem_116217951_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def get_item(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.getItem_263244906_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema = self._key_cache.get(request.TableName, tube)

            logger.debug('Sending GetItem request')
            Stubs.write_getItem_263244906_1(request, tube)

            result = self._decode_result('GetItem', request, Assemblers.getItem_263244906_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def put_item(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.putItem_N2106490455_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema = self._key_cache.get(request.TableName, tube)
            request.attr_names = AttributeValueEncoder._get_canonical_attribute_list(request.Item, request.key_schema)
            request.attr_list_id = self._attr_list_id_cache.get(request.attr_names, tube)

            logger.debug('Sending PutItem request')
            Stubs.write_putItem_N2106490455_1(request, tube)

            result = self._decode_result('PutItem', request, Assemblers.putItem_N2106490455_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def update_item(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.updateItem_1425579023_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema = self._key_cache.get(request.TableName, tube)

            logger.debug('Sending UpdateItem request')
            Stubs.write_updateItem_1425579023_1(request, tube)

            result = self._decode_result('UpdateItem', request, Assemblers.updateItem_1425579023_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def delete_item(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.deleteItem_1013539361_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema = self._key_cache.get(request.TableName, tube)

            logger.debug('Sending DeleteItem request')
            Stubs.write_deleteItem_1013539361_1(request, tube)

            result = self._decode_result('DeleteItem', request, Assemblers.deleteItem_1013539361_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def query(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.query_N931250863_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema = self._key_cache.get(request.TableName, tube)

            logger.debug('Sending Query request')
            Stubs.write_query_N931250863_1(request, tube)

            result = self._decode_result('Query', request, Assemblers.query_N931250863_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def scan(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(kwargs, DaxMethodIds.scan_N1875390620_1_Id))
        with self._tube_pool.get() as tube:
            tube.reauth()
            request.key_schema = self._key_cache.get(request.TableName, tube)

            logger.debug('Sending Scan request')
            Stubs.write_scan_N1875390620_1(request, tube)

            result = self._decode_result('Scan', request, Assemblers.scan_N1875390620_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def _convert_to_transaction_get_items_format(self, request, tube):
        """
        Segregate request data into separate arrays which will further be used for encoding the request. At this point basic validation from
        SDK is already done.
        :param request:
        :param tube:
        :return:
        """
        table_names = []
        keys = []
        projection_expressions = []

        key_set_per_table = {}
        key_schema_per_table = {}

        for transact_get_item in request['TransactItems']:

            operation_detail = list(transact_get_item.items())[0][1]

            table_name = operation_detail['TableName']
            key = operation_detail['Key']
            projection_expression = operation_detail.get('ProjectionExpression')
            attr_names = operation_detail.get('ExpressionAttributeNames')

            RequestValidator.validate_expression(proj_expr=projection_expression, attr_names=attr_names)

            key_schema = self._key_cache.get(table_name, tube)
            key_schema_per_table[table_name] = key_schema

            RequestValidator.validate_key(key, key_schema)
            key_bytes = AttributeValueEncoder.encode_key(key, key_schema)
            key_set = key_set_per_table.get(table_name)
            if not key_set:
                key_set = set()
                key_set_per_table[table_name] = key_set
            if key_bytes in key_set:
                raise RequestValidator.new_validation_exception("Transaction request cannot include multiple operations on one item")
            key_set.add(key_bytes)

            table_names.append(table_name.encode('utf-8'))
            keys.append(key_bytes)
            if projection_expression:
                encoded_expr = AttributeValueEncoder._parse_expressions_only(operation_detail)
                operation_detail['projection_ordinals'] = AttributeValueEncoder._prepare_projection(projection_expression,
                                                                                                    encoded_expr.ExpressionAttributeNames)
                projection_expression = encoded_expr.Projection
            projection_expressions.append(projection_expression)

        request.key_schema_by_table = key_schema_per_table
        request['tableNames'] = table_names
        request['keys'] = keys
        request['projectionExpressions'] = projection_expressions
        return request

    def transact_get_items(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(deepcopy(kwargs), DaxMethodIds.transactGetItems_1866287579_1_Id))

        with self._tube_pool.get() as tube:
            tube.reauth()

            request = self._convert_to_transaction_get_items_format(request, tube)

            logger.debug('Sending transact get items request')
            Stubs.write_transactGetItems_1866287579_1(request, tube)

            result = self._decode_result('TransactGetItems', request, Assemblers.transactGetItems_1866287579_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def _convert_to_transaction_write_items_format(self, request, tube):
        """
        Segregate request data into separate arrays which will further be used for encoding the request. At this point basic validation from
        SDK is already done.
        :param request:
        :param tube:
        :return:
        """
        operations = []
        table_names = []
        keys = []
        values = []
        return_values = None  # currently not used but expected by decoder on server side.
        return_values_on_condition_check_failures = []
        condition_expressions = []
        update_expressions = []

        key_set_per_table = {}
        key_schema_per_table = {}

        for transact_write_item in request['TransactItems']:
            key = None
            item = None

            operation_count = len(transact_write_item)
            if operation_count > 1:
                raise RequestValidator.new_validation_exception(
                    'TransactItems can only contain one of ConditionCheck, Put, Update or Delete')
            if operation_count == 0:
                raise RequestValidator.new_validation_exception(
                    "Invalid Request: TransactWriteRequest should contain Delete or Put or Update request")

            operation, operation_detail = list(transact_write_item.items())[0]
            if operation not in ['ConditionCheck', 'Put', 'Delete', 'Update']:
                raise RequestValidator.new_validation_exception(
                    "Invalid Request: TransactWriteRequest should contain Delete or Put or Update request")

            table_name = operation_detail['TableName']
            condition_expression = operation_detail.get('ConditionExpression')
            update_expression = operation_detail.get('UpdateExpression')
            attr_names = operation_detail.get('ExpressionAttributeNames')
            attr_vals = operation_detail.get('ExpressionAttributeValues')

            if operation == "ConditionCheck":
                operation = DynamoDBOperationNumerals.CHECK
                key = operation_detail['Key']
            if operation == "Put":
                operation_detail = transact_write_item["Put"]
                operation = DynamoDBOperationNumerals.PUT
                item = operation_detail['Item']
            if operation == "Delete":
                operation = DynamoDBOperationNumerals.DELETE
                key = operation_detail['Key']
            if operation == "Update":
                operation = DynamoDBOperationNumerals.PARTIAL_UPDATE
                key = operation_detail['Key']

            RequestValidator.validate_expression(cond_expr=condition_expression, upd_expr=update_expression, attr_names=attr_names,
                                                 attr_vals=attr_vals)
            # validate and encode key / value
            key_schema = self._key_cache.get(table_name, tube)
            key_schema_per_table[table_name] = key_schema
            if key:
                RequestValidator.validate_key(key, key_schema)
            key_bytes = AttributeValueEncoder.encode_key(key if key else item, key_schema)
            key_set = key_set_per_table.get(table_name)
            if not key_set:
                key_set = set()
                key_set_per_table[table_name] = key_set
            if key_bytes in key_set:
                raise RequestValidator.new_validation_exception("Transaction request cannot include multiple operations on one item")
            key_set.add(key_bytes)
            #  write to final keys array
            keys.append(key_bytes)

            if item:
                attr_names = AttributeValueEncoder._get_canonical_attribute_list(item, key_schema)
                attrListId = self._attr_list_id_cache.get(attr_names, tube)
                values.append(AttributeValueEncoder.encode_values(item, key_schema, attr_names, attrListId))
            else:
                values.append(None)

            operations.append(operation)
            table_names.append(table_name.encode('utf-8'))
            return_values_on_condition_check_failure = ReturnValuesOnConditionCheckFailure.NONE
            if operation_detail.get('ReturnValuesOnConditionCheckFailure') == 'ALL_OLD':
                return_values_on_condition_check_failure = ReturnValuesOnConditionCheckFailure.ALL_OLD
            return_values_on_condition_check_failures.append(return_values_on_condition_check_failure)

            # encode expressions along with attribute name and values
            encoded_expr = AttributeValueEncoder._parse_expressions_only(operation_detail)
            if encoded_expr.Condition:
                condition_expressions.append(encoded_expr.Condition)
            else:
                condition_expressions.append(None)
            if encoded_expr.Update:
                update_expressions.append(encoded_expr.Update)
            else:
                update_expressions.append(None)

        request.key_schema_by_table = key_schema_per_table
        request['operations'] = operations
        request['tableNames'] = table_names
        request['keys'] = keys
        request['values'] = values
        request['returnValues'] = return_values
        request['returnValuesOnConditionCheckFailure'] = return_values_on_condition_check_failures
        request['conditionExpressions'] = condition_expressions
        request['updateExpressions'] = update_expressions
        return request

    def transact_write_items(self, **kwargs):
        request = Request(DynamoDbV1Converter.convert_request(deepcopy(kwargs), DaxMethodIds.transactWriteItems_N1160037738_1_Id))

        with self._tube_pool.get() as tube:
            tube.reauth()

            request = self._convert_to_transaction_write_items_format(request, tube)

            logger.debug('Sending transact write items request')
            Stubs.write_transactWriteItems_N1160037738_1(request, tube)

            result = self._decode_result('TransactWriteItems', request, Assemblers.transactWriteItems_N1160037738_1, tube)
            result = self._resolve_attribute_values(result, tube)
            return result

    def endpoints(self):
        with self._tube_pool.get() as tube:
            tube.reauth()

            logger.debug('Sending endpoints request')
            Stubs.write_endpoints_455855874_1(tube)

            return self._decode_result('endpoints', None, Assemblers.endpoints_455855874_1, tube)

    def _defineKeySchema(self, table_name, tube):
        logger.debug('Sending defineKeySchema request')
        Stubs.write_defineKeySchema_N742646399_1(table_name, tube)

        return self._decode_result('defineKeySchema', None, Assemblers.defineKeySchema_N742646399_1, tube)

    def _defineAttributeList(self, attr_list_id, tube):
        if attr_list_id == 1:
            # 1 is defined to be an empty array
            return []

        logger.debug('Sending defineAttributeList request')
        Stubs.write_defineAttributeList_670678385_1(attr_list_id, tube)

        return self._decode_result('defineAttributeList', None, Assemblers.defineAttributeList_670678385_1, tube)

    def _defineAttributeListId(self, attr_list, tube):
        if len(attr_list) == 0:
            # An empty list is defined to be 1
            return 1

        logger.debug('Sending defineAttributeListId request')
        Stubs.write_defineAttributeListId_N1230579644_1(attr_list, tube)

        return self._decode_result('defineAttributeListId', None, Assemblers.defineAttributeListId_N1230579644_1, tube)

    def _decode_result(self, operation_name, request, assembler, tube):
        logger.debug('Decoding %s response', operation_name)
        try:
            status = tube.peek()
        except CborDecoder.NoMoreData as e:
            raise DaxServiceError.from_no_more_data_exception(operation_name, e)

        if status == DaxClient.SUCCESS:
            tube.skip()  # Throw away the empty error header
            return assembler(request, tube)
        else:
            return self._handle_error(operation_name, tube)

    def _handle_error(self, operation_name, tube):
        codes = tube.read_object()
        message = tube.read_string()
        exc_info = tube.read_object() or (None, None, None)
        raise DaxServiceError(operation_name, message, codes, *exc_info)

    def _resolve_attribute_values(self, result, tube):
        if 'Item' in result:
            self._resolve_item_attribute_values(result['Item'], tube)

        if 'Attributes' in result:
            self._resolve_item_attribute_values(result['Attributes'], tube)

        if 'Items' in result:
            for item in result['Items']:
                self._resolve_item_attribute_values(item, tube)

        if 'Responses' in result:
            if isinstance(result['Responses'], list):
                # transact-get-items
                for response in result['Responses']:
                    # skip the empty responses
                    if response:
                        self._resolve_item_attribute_values(response['Item'], tube)
            else:
                # batch-get-items
                for table_results in result['Responses'].values():
                    for item in table_results:
                        self._resolve_item_attribute_values(item, tube)

        if 'UnprocessedItems' in result:
            for unprocessed_requests in result['UnprocessedItems'].values():
                for unprocessed_request in unprocessed_requests:
                    for write_type, write_request in unprocessed_request.items():
                        if write_type == 'PutRequest':
                            self._resolve_item_attribute_values(write_request['Item'], tube)

        return result

    def _resolve_item_attribute_values(self, item, tube):
        if '_attr_list_id' in item:
            attr_names = self._attr_list_cache.get(item['_attr_list_id'], tube)
            AttributeValueDecoder.deanonymize_attribute_values(item, attr_names)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import re
import json
from collections import defaultdict

from .DaxError import DaxValidationError
from .CborEncoder import CborEncoder, DYNAMODB_CONTEXT
from .Constants import DaxDataRequestParam, BINARY_TYPES
from .DocumentPath import DocumentPath
from . import LexDecimal, DaxCborTypes
from .compat import to_bytes

def encode_key(item, schema):
    enc = CborEncoder()
    encode_key_direct(enc, item, schema)
    return enc.as_bytes()

def encode_key_direct(enc, item, schema):
    hashAttrDefn = schema[0]
    hashAttrVal = item.get(hashAttrDefn['AttributeName'])
    if not hashAttrVal:
        raise DaxValidationError('One of the required keys was not given a value.')

    hashAttrType = hashAttrDefn['AttributeType']
    hashVal = hashAttrVal.get(hashAttrType)
    if not hashVal:
        raise DaxValidationError('One of the required keys was not given a value.')

    if len(schema) == 2:
        if hashAttrType == 'S':
            enc.append_string(hashVal)
        elif hashAttrType == 'N':
            enc.append_number(hashVal)
        elif hashAttrType == 'B':
            enc.append_binary(hashVal)
        else:
            raise DaxValidationError('Unsupported KeyType encountered in Hash Attribute: ' + str(hashAttrType))

        rangeAttrDefn = schema[1]
        rangeAttrVal = item.get(rangeAttrDefn['AttributeName'])
        if not rangeAttrVal:
            raise DaxValidationError('One of the required keys was not given a value.')

        rangeAttrType = rangeAttrDefn['AttributeType']
        rangeVal = rangeAttrVal.get(rangeAttrType)
        if not rangeVal:
            raise DaxValidationError('One of the required keys was not given a value.')
        if rangeAttrType == 'S':
            enc.append_raw(to_bytes(rangeVal, 'utf8'))
        elif rangeAttrType == 'N':
            enc.append_raw(LexDecimal.encode(DYNAMODB_CONTEXT.create_decimal(rangeVal)))
        elif rangeAttrType == 'B':
            enc.append_raw(rangeVal)
        else:
            raise DaxValidationError('Unsupported KeyType encountered in Range Attribute: ' + str(rangeAttrType))

    elif len(schema) == 1:
        if hashAttrType == 'S':
            enc.append_raw(to_bytes(hashVal, 'utf8'))
        elif hashAttrType == 'N':
            enc.append_number(hashVal)
        elif hashAttrType == 'B':
            enc.append_raw(hashVal)
        else:
            raise DaxValidationError('Unsupported KeyType encountered in Hash Attribute: ' + str(hashAttrType))

    else:
        raise DaxValidationError("Unexpected key length: " + str(len(schema)))

def encode_compound_key(key):
    enc = CborEncoder()
    encode_compound_key_direct(enc, key)
    return enc.as_bytes()

def encode_compound_key_direct(enc, key):
    # This must be a map stream or the server will error
    enc.append_map_stream_header()

    for name, av in key.items():
        enc.append_string(name)
        encode_attribute_value_direct(enc, av)

    enc.append_break()

def encode_values(item, key_schema, attr_names, attr_list_id):
    enc = CborEncoder()

    attr_names = attr_names or _get_canonical_attribute_list(item, key_schema)
    encode_attributes_direct(enc, item, attr_names, attr_list_id)

    return enc.as_bytes()

def encode_attributes_direct(enc, item, attr_names, attr_list_id):
    enc.append_int(attr_list_id)
    for attr in attr_names:
        av = item[attr]
        encode_attribute_value_direct(enc, av)

def encode_attribute_value(av):
    enc = CborEncoder()
    encode_attribute_value_direct(enc, av)
    return enc.as_bytes()

def encode_attribute_value_direct(enc, av):
    if len(av) != 1:
        raise DaxValidationError('Attribute Values must have exactly one key, got: ' + str(list(av.keys())))

    # Yes, this loop will only execute once, but it's the easiest way to get at the value...
    for t, value in av.items():
        if t == 'S':
            enc.append_string(value)
        elif t == 'N':
            enc.append_number(value)
        elif t == 'B':
            enc.append_binary(value)
        elif t == 'SS':
            if len(value) == 0:
                raise DaxValidationError('Supplied AttributeValue is empty, must contain exactly one of the supported datatypes')
            enc.append_tag(DaxCborTypes.TAG_DDB_STRING_SET)
            enc.append_array_header(len(value))
            for s in value:
                enc.append_string(s)
        elif t == 'NS':
            if len(value) == 0:
                raise DaxValidationError('Supplied AttributeValue is empty, must contain exactly one of the supported datatypes')
            enc.append_tag(DaxCborTypes.TAG_DDB_NUMBER_SET)
            enc.append_array_header(len(value))
            for n in value:
                enc.append_number(n)
        elif t == 'BS':
            if len(value) == 0:
                raise DaxValidationError('Supplied AttributeValue is empty, must contain exactly one of the supported datatypes')
            enc.append_tag(DaxCborTypes.TAG_DDB_BINARY_SET)
            enc.append_array_header(len(value))
            for b in value:
                enc.append_binary(b)
        elif t == 'M':
            enc.append_map_header(len(value))
            for key, sub_av in sorted(value.items()):
                enc.append_string(key)
                encode_attribute_value_direct(enc, sub_av)
        elif t == 'L':
            enc.append_array_header(len(value))
            for sub_av in value:
                encode_attribute_value_direct(enc, sub_av)
        elif t == 'NULL':
            enc.append_null()
        elif t == 'BOOL':
            enc.append_boolean(value)
        else:
            raise DaxValidationError("Unknown Attribute Value type '{}'".format(t))

        break # Ensure that it only executes once

def encode_expressions(request):
    enc = CborEncoder()
    encode_expressions_direct(enc, request)
    return enc.as_bytes()

def encode_key_condition_expression_direct(enc, request):
    expressions = _parse_expressions_only(request)

    if 'KeyConditionExpression' in request:
        enc.append_binary(expressions.KeyCondition)
    else:
        enc.append_null()

    return expressions

def encode_expressions_direct(enc, request):
    expressions = _parse_expressions_only(request)
    write_expressions_direct(enc, expressions, request)

def write_expressions_direct(enc, expressions, request):
    if expressions.Condition:
        enc.append_int(DaxDataRequestParam.ConditionExpression)
        enc.append_binary(expressions.Condition)

    if expressions.Filter:
        enc.append_int(DaxDataRequestParam.FilterExpression)
        enc.append_binary(expressions.Filter)

    if expressions.Update:
        enc.append_int(DaxDataRequestParam.UpdateExpression)
        enc.append_binary(expressions.Update)

    if expressions.Projection:
        # Store the projection ordinals for use reading the response item
        request['_projection_ordinals'] = _prepare_projection(
                request.ProjectionExpression, expressions.ExpressionAttributeNames)

        enc.append_int(DaxDataRequestParam.ProjectionExpression)
        enc.append_binary(expressions.Projection)

    if expressions.ExpressionAttributeNames:
        enc.append_int(DaxDataRequestParam.ExpressionAttributeNames)
        enc.append_map_header(len(expressions.ExpressionAttributeNames))
        for k, v in sorted(expressions.ExpressionAttributeNames.items()):
            enc.append_string(k)
            enc.append_string(v)

    if expressions.ExpressionAttributeValues:
        enc.append_int(DaxDataRequestParam.ExpressionAttributeValues)
        enc.append_map_header(len(expressions.ExpressionAttributeValues))
        for k, v in sorted(expressions.ExpressionAttributeValues.items()):
            enc.append_string(k)
            encode_attribute_value_direct(enc, v)

def encode_ExclusiveStartKey_direct(enc, request):
    if 'IndexName' in request:
        enc.append_binary(encode_compound_key(request.ExclusiveStartKey))
    else:
        enc.append_binary(encode_key(request.ExclusiveStartKey, request.key_schema))

def encode_batchGetItem_N697851100_1_getItemKeys_direct(enc, request):
    # _validate_batchGetItem(request)
    table_proj_ordinals = defaultdict(list)
    key_set = set()

    enc.append_map_header(len(request.RequestItems))

    for table_name, table_request in request.RequestItems.items():
        enc.append_string(table_name)
        enc.append_array_header(3)

        # Item 1: ConsistentRead
        consistent_read = table_request.get('ConsistentRead', False)
        enc.append_boolean(consistent_read)

        # Item 2: Projection Expression
        proj_expression = table_request.get('ProjectionExpression')
        if proj_expression is not None:
            expressions = _parse_expressions_only(table_request)
            table_proj_ordinals[table_name] = _prepare_projection(
                    proj_expression, expressions.ExpressionAttributeNames)
            enc.append_binary(expressions.Projection)
        else:
            enc.append_null()

        keys = table_request.get('Keys', [])

        # Item 3: Keys array
        enc.append_array_header(len(keys))
        for key in keys:
            key_bytes = encode_key(key, request.key_schema_by_table[table_name])
            if key_bytes in key_set:
                raise DaxValidationError('Provided list of item keys contains duplicates:' + json.dumps(key))

            key_set.add(key_bytes)
            enc.append_binary(key_bytes)

    # Store on the request for decoding
    request._projection_ordinals_by_table = dict(table_proj_ordinals)

WRITE_TYPES = {'PutRequest', 'DeleteRequest'}
MAX_WRITE_BATCH_SIZE = 25

def encode_batchWriteItem_116217951_1_keyValuesByTable_direct(enc, request):
    # TODO Run all validation prior to writing anything to the stream
    enc.append_map_header(len(request.RequestItems))

    key_set = set()
    total_requests = 0
    for table_name, write_requests in request.RequestItems.items():
        key_schema = request.key_schema_by_table[table_name]
        key_set.clear()

        total_requests += len(write_requests)
        if total_requests > MAX_WRITE_BATCH_SIZE:
            raise DaxValidationError(
                    "Batch size should be less than {}, got {}".format(MAX_WRITE_BATCH_SIZE, total_requests))

        enc.append_string(table_name)

        request_item_count = 0
        is_empty = True
        request_item_count = sum(sum(1 for write_type in write_request if write_type in WRITE_TYPES) \
                for write_request in write_requests)

        enc.append_array_header(request_item_count * 2)

        for write_request in write_requests:
            is_empty = False
            if 'PutRequest' in write_request:
                put_request = write_request['PutRequest']
                item = put_request['Item']
                key_bytes = encode_key(item, key_schema)
                if key_bytes in key_set:
                    raise DaxValidationError('Provided list of item keys contains duplicates:' + json.dumps(item))

                key_set.add(key_bytes)

                enc.append_binary(key_bytes)
                enc.append_binary(
                        encode_values(item, key_schema, put_request['attr_names'], put_request['attr_list_id']))

            if 'DeleteRequest' in write_request:
                del_request = write_request['DeleteRequest']
                key = del_request['Key']
                key_bytes = encode_key(key, request.key_schema_by_table[table_name])
                if key_bytes in key_set:
                    raise DaxValidationError('Provided list of item keys contains duplicates:' + json.dumps(key))

                key_set.add(key_bytes)

                enc.append_binary(key_bytes)
                enc.append_null()

        if is_empty:
            raise DaxValidationError("No write requests provided for " + table_name)

    if total_requests == 0:
        raise DaxValidationError("No write requests provided")

def _parse_expressions_only(request):
    # import here to avoid circular reference
    from . import CborSExprGenerator

    expr_attr_names = request.get('ExpressionAttributeNames', {})
    expr_attr_values = request.get('ExpressionAttributeValues', {})

    _check_valid_expression_parameter_names(expr_attr_names, expr_attr_values)

    expressions = CborSExprGenerator.encode_expressions(
        request.get('ConditionExpression'),
        request.get('KeyConditionExpression'),
        request.get('FilterExpression'),
        request.get('UpdateExpression'),
        request.get('ProjectionExpression'),
        expr_attr_names,
        expr_attr_values)

    return expressions

def _check_valid_expression_parameter_names(expr_attr_names, expr_attr_values):
    if expr_attr_names:
        invalid = _check_expression_params(expr_attr_names.keys(), '#')
        if invalid:
            raise DaxValidationError('ExpressionAttributeNames contains invalid key: "' + invalid + '"')

    if expr_attr_values:
        invalid = _check_expression_params(expr_attr_values.keys(), ':')
        if invalid:
            raise DaxValidationError('ExpressionAttributeValues contains invalid key: "' + invalid + '"')

_EXPRESSION_PARAM_NAME_RX = re.compile(r'^[#:][A-Za-z0-9_]+$')
def _check_expression_params(key_names, prefix):
    if not key_names:
        return

    for name in key_names:
        if name[0] != prefix or not _EXPRESSION_PARAM_NAME_RX.match(name):
            return name

def _prepare_projection(expr, expr_attr_names):
    return [DocumentPath.from_path(path.strip(), expr_attr_names) for path in expr.split(',')]

def _get_canonical_attribute_list(item, key_schema):
    key_names = {key['AttributeName'] for key in key_schema}
    return tuple(_as_string(attr_name) for attr_name in sorted(set(item.keys()) - key_names))

def _as_string(s):
    ''' Convert a byte string to a text string if necessary.

    DAX requires the attribute list to be Strings, and the default encoder under Python 2 sends str as a byte array.

    This is probably going to cause errors somewhere, but such is life in Python 2. It uses the default encoding
    as these strings are most likely from the source code.
    '''
    return s.decode() if isinstance(s, BINARY_TYPES) else s



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import decimal
import six

STRING_TYPES = (six.text_type,)
BINARY_TYPES = (bytes, bytearray)
INTEGRAL_TYPES = six.integer_types
REAL_TYPES = (float, decimal.Decimal)
NUMBER_TYPES = INTEGRAL_TYPES + REAL_TYPES
SEQ_TYPES = (list, tuple)

PY_TO_OP_NAME = {
    'get_item': 'GetItem',
    'put_item': 'PutItem',
    'delete_item': 'DeleteItem',
    'update_item': 'UpdateItem',
    'query': 'Query',
    'scan': 'Scan',
    'batch_get_item': 'BatchGetItem',
    'batch_write_item': 'BatchWriteItem',
    'transact_get_items': 'TransactGetItems',
    'transact_write_items': 'TransactWriteItems'
}


class DaxMethodIds:
    authorizeConnection_1489122155_1_Id = 1489122155
    batchGetItem_N697851100_1_Id = -697851100
    batchWriteItem_116217951_1_Id = 116217951
    defineAttributeList_670678385_1_Id = 670678385
    defineAttributeListId_N1230579644_1_Id = -1230579644
    defineKeySchema_N742646399_1_Id = -742646399
    deleteItem_1013539361_1_Id = 1013539361
    endpoints_455855874_1_Id = 455855874
    getItem_263244906_1_Id = 263244906
    methods_785068263_1_Id = 785068263
    putItem_N2106490455_1_Id = -2106490455
    query_N931250863_1_Id = -931250863
    scan_N1875390620_1_Id = -1875390620
    services_N1016793520_1_Id = -1016793520
    updateItem_1425579023_1_Id = 1425579023
    transactGetItems_1866287579_1_Id = 1866287579
    transactWriteItems_N1160037738_1_Id = -1160037738


class DaxResponseParam:
    Item = 0
    ConsumedCapacity = 1
    Attributes = 2
    ItemCollectionMetrics = 3
    Responses = 4
    UnprocessedKeys = 5
    UnprocessedItems = 6
    Items = 7
    Count = 8
    LastEvaluatedKey = 9
    ScannedCount = 10
    TableDescription = 11


class DaxDataRequestParam:
    ProjectionExpression = 0
    ExpressionAttributeNames = 1
    ConsistentRead = 2
    ReturnConsumedCapacity = 3
    ConditionExpression = 4
    ExpressionAttributeValues = 5
    ReturnItemCollectionMetrics = 6
    ReturnValues = 7
    UpdateExpression = 8
    ExclusiveStartKey = 9
    FilterExpression = 10
    IndexName = 11
    KeyConditionExpression = 12
    Limit = 13
    ScanIndexForward = 14
    Select = 15
    Segment = 16
    TotalSegments = 17
    RequestItems = 18
    ClientRequestToken = 19


class ReturnConsumedCapacityValues:
    NONE = 0
    TOTAL = 1
    INDEXES = 2


class ReturnItemCollectionMetricsValue:
    NONE = 0
    SIZE = 1


class SelectValues:
    ALL_ATTRIBUTES = 1
    ALL_PROJECTED_ATTRIBUTES = 2
    COUNT = 3
    SPECIFIC_ATTRIBUTES = 4


class ReturnValues:
    NONE = 1
    ALL_OLD = 2
    UPDATED_OLD = 3
    ALL_NEW = 4
    UPDATED_NEW = 5


class ReturnValuesValues:
    NONE = 1
    ALL_OLD = 2
    UPDATED_OLD = 3
    ALL_NEW = 4
    UPDATED_NEW = 5


class ReturnItemCollectionMetricsValues:
    NONE = 0
    SIZE = 1


CONSUMED_CAPACITY = {
    "CapacityUnits": 1,
    "ReadCapacityUnits": 2,
    "WriteCapacityUnits": 3,
    "TableName": 4,
    "Table": 5,
    "GlobalSecondaryIndexes": 6,
    "LocalSecondaryIndexes": 7
}


class DynamoDBOperationNumerals:
    GET = 1
    PUT = 2
    EXCHANGE = 3
    INSERT = 4
    REPLACE = 5
    UPDATE = 6
    DELETE = 7
    REMOVE = 8
    PARTIAL_UPDATE = 9
    BATCH_GET = 10
    BATCH_WRITE = 11
    CHECK = 12
    TRANSACT_WRITE = 13
    TRANSACT_GET = 14
    SCAN = 15
    QUERY = 16
    CREATE_TABLE = 17
    DELETE_TABLE = 18
    DESCRIBE_TABLE = 19
    LIST_TABLES = 20
    UPDATE_TABLE = 21


class ReturnValuesOnConditionCheckFailure:
    NONE = 1
    ALL_OLD = 2


# Default base sleep time (milliseconds) for non-throttled exceptions.
# https://github.com/aws/aws-sdk-java/blob/7b1e5b87b0bf03456df9e77716b14731adf9a7a7/aws-java-sdk-core/src/main/java/com/amazonaws/retry/PredefinedBackoffStrategies.java#L33
SDK_DEFAULT_BASE_DELAY_MS = 100

# Default base sleep time (milliseconds) for throttled exceptions.
# https://github.com/aws/aws-sdk-java/blob/7b1e5b87b0bf03456df9e77716b14731adf9a7a7/aws-java-sdk-core/src/main/java/com/amazonaws/retry/PredefinedBackoffStrategies.java#L38
SDK_DEFAULT_THROTTLED_BASE_DELAY_MS = 500



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# TODO: Better buffer management - track pointers in the buffer rather than
# trimming the front

import struct
import binascii
import io

from collections import OrderedDict

from .CborTypes import * # pylint: disable=wildcard-import,unused-wildcard-import
from .CborEncoder import DYNAMODB_CONTEXT
from .DaxError import DaxClientError, DaxErrorCode

class NoMoreData(Exception):
    pass

class CborDecoder(object):
    ''' Decode CBOR from a stream.

    Data is provided by the `more` function, which is:

      more(buf: buffer, n: int) -> int

    It should add at lest n bytes to the buffer from whatever source. Returns the number of bytes requested,
    or -1 if no more bytes are available and never will be - socket closed, EOF, etc.

    TODO Buffer handling is very naive, as it just resizes the bytearray. Better
    would be to track locations in the buffer to avoid reallocating.

    TODO Ignore unexpected tags during all decoding.

    TODO Add chunked reads of fixed-size byte records (decode_bytes_iter?)
    '''
    def __init__(self, more, tag_handlers=None):
        self.more = more
        self.buffer = bytearray()

        self.tag_handlers = _DEFAULT_TAG_HANDLERS.copy()
        if tag_handlers:
            self.tag_handlers.update(tag_handlers)

    def peek(self):
        self._ensure_available(1)
        return self.buffer[0]

    def skip(self):
        # skip to the next entry
        # TODO Do this more efficiently, without creating an object and just reading types
        self.decode_object()

    def _consume(self, n=1):
        del self.buffer[:n]

    def drain(self):
        ''' Return all available remaining bytes.

        This assumes a fixed amount of data (as in a nested decoder) or a
        suitable end-of-stream, otherwise it may try to run forever.
        '''
        while True:
            try:
                self._ensure_available(4096)
            except NoMoreData:
                break

        result = bytes(self.buffer)
        del self.buffer[:]
        return result

    def decode_string(self):
        t = self.peek()
        if not is_major_type(t, TYPE_UTF):
            raise DaxClientError('Type is not string: ' + major_type_name(t), DaxErrorCode.Decoder)

        length = self._decode_value(t)

        if length != -1:
            self._ensure_available(length)
            result = self.buffer[:length].decode('utf-8')
            self._consume(length)
        else:
            f = io.StringIO()
            while not self._try_decode_break():
                f.write(self.decode_string())
            result = f.getvalue()

        return result

    def decode_binary(self):
        t = self.peek()
        if not is_major_type(t, TYPE_BYTES):
            raise DaxClientError('Type is not bytes: ' + major_type_name(t), DaxErrorCode.Decoder)

        length = self._decode_value(t)

        if length != -1:
            self._ensure_available(length)
            result = self.buffer[:length]
            self._consume(length)
        else:
            f = io.BytesIO()
            while not self._try_decode_break():
                f.write(self.decode_binary())
            result = f.getvalue()

        return bytes(result)

    def decode_number(self):
        t = self.peek()
        if t in (TYPE_FLOAT_16, TYPE_FLOAT_32, TYPE_FLOAT_64):
            return self.decode_float()

        mt = major_type(t)
        if mt in (TYPE_POSINT, TYPE_NEGINT):
            return self.decode_int()
        elif mt == TYPE_TAG:
            tag = self._decode_tag(t)
            if tag in (TAG_POSBIGINT, TAG_NEGBIGINT):
                return _decode_bigint(self, tag)
            elif tag == TAG_DECIMAL:
                return _decode_decimal(self, tag)

        raise DaxClientError('Type is not a number: ' + major_type_name(t), DaxErrorCode.Decoder)

    def decode_int(self):
        t = self.peek()

        if major_type(t) not in (TYPE_POSINT, TYPE_NEGINT):
            if is_major_type(t, TYPE_TAG):
                tag = self._decode_tag(t)
                if tag in (TAG_POSBIGINT, TAG_NEGBIGINT):
                    return _decode_bigint(self, tag)

            raise DaxClientError('Type is not int: ' + major_type_name(t), DaxErrorCode.Decoder)

        v = self._decode_value(t)
        return v if is_major_type(t, TYPE_POSINT) else -v - 1

    def decode_decimal(self):
        t = self.peek()
        if not is_major_type(t, TYPE_TAG):
            raise DaxClientError('Decimal value must have TAG_DECIMAL tag.', DaxErrorCode.Decoder)

        tag = self._decode_tag(t)
        return _decode_decimal(self, tag)

    def decode_float(self):
        t = self.peek()

        if t == TYPE_FLOAT_16:
            self._ensure_available(3)
            result = _parseHalf(self.buffer[1:3])
            excess = 2
        elif t == TYPE_FLOAT_32:
            self._ensure_available(5)
            result, = struct.unpack_from('!f', self.buffer, 1)
            excess = 4
        elif t == TYPE_FLOAT_64:
            self._ensure_available(9)
            result, = struct.unpack_from('!d', self.buffer, 1)
            excess = 8
        else:
            raise DaxClientError('Type is not float: ' + major_type_name(t), DaxErrorCode.Decoder)

        self._consume(1+excess)
        return result

    def decode_map(self):
        # A simple dict comprehension doesn't work here, because the order of call _dec.decode_object() doesn't work
        # The calls need to be split into seperate variables
        def _():
            for _dec in self.decode_map_iter():
                k = _dec.decode_object()
                v = _dec.decode_object()
                yield k, v

        return OrderedDict(_())

    def decode_map_iter(self):
        ''' Allow for iterative decoding of maps, rather than reading the whole map in.

        This function returns a generator that produces CborDecoder instances that can be used to read the next
        elements. It handles both fixed-length and streaming maps.

        Example:
            for _dec in dec.read_map_iter():
                key = _dec.decode_object()
                value = _dec.decode_object()
                result[key] = value
        '''
        size = self._decode_map_header()

        i = 0
        while i != size:
            if self._try_decode_break():
                break

            yield self
            i += 1

    def decode_array(self):
        return [_dec.decode_object() for _dec in self.decode_array_iter()]

    def decode_array_iter(self):
        ''' Allow for iterative decoding of arrays, rather than reading the whole array in.

        This function returns a generator that produces CborDecoder instances that can be used to read the next
        elements. It handles both fixed-length and streaming arrays.

        Example:
            for _dec in dec.read_array_iter():
                value = _dec.decode_object()
                result.append(value)
        '''
        size = self._decode_array_header()

        i = 0
        while i != size:
            if self._try_decode_break():
                break

            yield self
            i += 1

    def decode_object(self):
        if self.try_decode_null():
            return None

        t = self.peek()

        # Check simple types first
        if t == TYPE_NULL or t == TYPE_UNDEFINED:
            self._consume(1)
            return None
        elif t == TYPE_TRUE:
            self._consume(1)
            return True
        elif t == TYPE_FALSE:
            self._consume(1)
            return False
        elif t in (TYPE_FLOAT_16, TYPE_FLOAT_32, TYPE_FLOAT_64):
            return self.decode_float()
        elif t == TYPE_BREAK:
            raise DaxClientError('Unexpected break', DaxErrorCode.Decoder)
        else:
            # proceed to complex types
            pass

        mt = major_type(t)
        if mt == TYPE_POSINT or mt == TYPE_NEGINT:
            return self.decode_int()
        elif mt == TYPE_BYTES:
            return self.decode_binary()
        elif mt == TYPE_UTF:
            return self.decode_string()
        elif mt == TYPE_ARRAY:
            return self.decode_array()
        elif mt == TYPE_MAP:
            return self.decode_map()
        elif mt == TYPE_TAG:
            return self._decode_tagged_type(t)
        else:
            raise DaxClientError('Unhandled type: ' + major_type_name(t), DaxErrorCode.Decoder)

    def decode_cbor(self):
        ''' Read nested CBOR stored in a bytes record.

        Returns a CborDecoder instance that can be used to read the nested CBOR.

        TODO Used a chunked reading (decode_binary_iter) when available, or tie the nested more function into the
        parent more, rather than reading the whole bytes record.
        '''
        data = self.decode_binary()

        def more(buf, n):
            # Normally would use nonlocal if 2.x support not needed
            if more.p >= len(data):
                return -1
            q = more.p+n
            buf += data[more.p:q]
            more.p = q
            return n
        more.p = 0

        dec = self.__class__(more)
        return dec

    def _decode_map_header(self):
        t = self.peek()
        if not is_major_type(t, TYPE_MAP):
            raise DaxClientError('Type is not map (was {})'.format(major_type_name(t)), DaxErrorCode.Decoder)

        return self._decode_value(t)

    def _decode_array_header(self):
        t = self.peek()
        if not is_major_type(t, TYPE_ARRAY):
            raise DaxClientError('Type is not array (was {})'.format(major_type_name(t)), DaxErrorCode.Decoder)

        return self._decode_value(t)

    def _decode_tag(self, t):
        if not is_major_type(t, TYPE_TAG):
            raise DaxClientError('Type is not a tag (was {})'.format(major_type_name(t)), DaxErrorCode.Decoder)

        return self._decode_value(t)

    def _decode_tagged_type(self, t):
        tag = self._decode_tag(t)
        if tag in self.tag_handlers:
            return self.tag_handlers[tag](self, tag)
        else:
            raise DaxClientError('Unknown tag ' + str(tag), DaxErrorCode.Decoder)

    def _try_decode_break(self):
        val = self.peek()

        if val == TYPE_BREAK:
            self._consume()
            return True
        else:
            return False

    def try_decode_null(self):
        val = self.peek()

        if val == TYPE_NULL:
            self._consume()
            return True
        else:
            return False

    def _decode_value(self, v):
        size = minor_type(v)
        if size < SIZE_8:
            result = size
            excess = 0
        elif size == SIZE_8:
            self._ensure_available(2)
            result = self.buffer[1]
            excess = 1
        elif size == SIZE_16:
            self._ensure_available(3)
            result, = struct.unpack_from("!H", self.buffer, 1)
            excess = 2
        elif size == SIZE_32:
            self._ensure_available(5)
            result, = struct.unpack_from("!I", self.buffer, 1)
            excess = 4
        elif size == SIZE_64:
            self._ensure_available(9)
            result, = struct.unpack_from("!Q", self.buffer, 1)
            excess = 8
        elif size == SIZE_STREAM:
            result = -1
            excess = 0
        else:
            raise DaxClientError('Invalid value size', DaxErrorCode.Decoder)

        self._consume(1+excess)
        return result

    def _ensure_available(self, n):
        ''' Ensure that at least n bytes are available in the buffer.

        This will call the `more` callback until the desired amount is reached.
        '''
        while len(self.buffer) < n:
            l = n - len(self.buffer)
            m = self.more(self.buffer, l)
            if m <= 0:
                raise NoMoreData()

def _decode_bigint(dec, tag):
    if tag not in (TAG_POSBIGINT, TAG_NEGBIGINT):
        raise DaxClientError('Invalid tag to decode BigInt: ' + hex(tag), DaxErrorCode.Decoder)

    t = dec.peek()
    if not is_major_type(t, TYPE_BYTES):
        raise DaxClientError('Type for BigInt is not binary: ' + major_type_name(t), DaxErrorCode.Decoder)

    data = dec.decode_binary()
    val = int(binascii.hexlify(data), 16)

    return val if tag == TAG_POSBIGINT else -val - 1

def _decode_decimal(dec, tag):
    if tag != TAG_DECIMAL:
        raise DaxClientError('Decimal value must have TAG_DECIMAL tag (got ' + hex(tag) + ')', DaxErrorCode.Decoder)

    size = dec._decode_array_header()
    if size != 2:
        raise DaxClientError('Decimal value has wrong array size (' + str(size) + ')', DaxErrorCode.Decoder)

    exponent = dec.decode_int()
    mantissa = dec.decode_int()

    # There should be a faster way to do this than string formatting
    if mantissa >= 0:
        sign = 0
        digits = tuple(map(int, str(mantissa)))
    else:
        sign = 1
        digits = tuple(map(int, str(-mantissa)))

    return DYNAMODB_CONTEXT.create_decimal((sign, digits, exponent))


_DEFAULT_TAG_HANDLERS = {
    TAG_POSBIGINT: _decode_bigint,
    TAG_NEGBIGINT: _decode_bigint,
    TAG_DECIMAL: _decode_decimal,
}

_NaN = float('NaN')

def _parseHalf(buf):
    sign = -1 if buf[0] & 0x80 else 1
    exp = (buf[0] & 0x7C) >> 2
    mant = ((buf[0] & 0x03) << 8) | buf[1]
    if not exp:
        return sign * 5.9604644775390625e-8 * mant
    elif exp == 0x1f:
        return sign * (_NaN if mant else 2e308)
    else:
        return sign * pow(2, exp - 25) * (1024 + mant)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import boto3

from .AmazonDaxClient import AmazonDaxClient

def is_boto3_session(session):
    return isinstance(session, boto3.session.Session)

class DaxSession(boto3.session.Session):
    @classmethod
    def from_boto3(cls, session, endpoints):
        ''' Create a DaxSession from a boto3 Session by copying the internals. '''
        # This makes a copy in case the session is reused in other contexts
        # Hopefully copying doesn't have any other side-effects
        return cls(session._session.credentials.aws_access_key_id,
            session._session.credentials.aws_secret_access_key,
            session._session.credentials.aws_session_token,
            session.region_name,
            session._session,
            session.profile_name,
            endpoints)

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 aws_session_token=None, region_name=None,
                 botocore_session=None, profile_name=None, endpoints=None):
        super(DaxSession, self).__init__(aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name)
        self.endpoints = endpoints

    def client(self, service_name, region_name=None, api_version=None,
               use_ssl=True, verify=None, endpoint_url=None,
               aws_access_key_id=None, aws_secret_access_key=None,
               aws_session_token=None, config=None):
        ''' Create a DAX client instead of a dynamodb client. '''
        if service_name != 'dynamodb' and service_name != 'dax':
            raise ValueError('service_name must be "dynamodb" or "dax"')

        return AmazonDaxClient(self,
            region_name=region_name, api_version=api_version,
            use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token, config=config,
            endpoints=self.endpoints)

    def resource(self, service_name, region_name=None, api_version=None,
                 use_ssl=True, verify=None, endpoint_url=None,
                 aws_access_key_id=None, aws_secret_access_key=None,
                 aws_session_token=None, config=None):
        ''' Create a DAX resource, which is really just a dynamodb resource with a different client. '''
        if service_name != 'dynamodb' and service_name != 'dax':
            raise ValueError('service_name must be "dynamodb" or "dax"')

        res = super(DaxSession, self).resource('dynamodb',
            region_name=region_name, api_version=api_version,
            use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token, config=config)

        # BLACK MAGIC (not really)
        # The context manager methods __enter__ and __exit__ must be defined on the type, not the instance
        # But what we have is an instance with a dynamically-generated base class, and sticking the methods on
        # that might propagate to other instances.
        # It looks like the current version of boto generates a new class on every call to .resource()
        # but I don't want to rely on that.
        # Soooo ... we generate a new dynamic class that uses THAT dynamic class as its base,
        # and stick __enter__ and __exit__ on it.
        dax_svc_res_attrs = {
            '__enter__': _resource_enter,
            '__exit__': _resource_exit
        }
        dax_svc_res = type('DaxServiceResource', (type(res),), dax_svc_res_attrs)

        # Then, make the new class the base class of the existing instance
        res.__class__ = dax_svc_res

        # Now the resource has our custom class as its base but all other attributes and behaviour are preserved

        return res

def _resource_enter(self):
    return self

def _resource_exit(self, exc_type, exc_value, traceback):
    return self.meta.client.__exit__(exc_type, exc_value, traceback)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from . import CborTypes, CborDecoder, LexDecimal
from .DaxCborTypes import DdbSet
from .DaxError import DaxClientError, DaxErrorCode
from .CborTypes import is_major_type, major_type
from .ItemBuilder import ItemBuilder
from .Constants import STRING_TYPES, BINARY_TYPES, NUMBER_TYPES

def deanonymize_attribute_values(item, attr_names):
    ''' Resolve any attribute lists into attribute value maps in the given item.'''
    attr_values = item['_anonymous_attribute_values']

    if isinstance(attr_values, dict):
        values = {attr_names[idx]: value for idx, value in attr_values.items()}
    else:
        if len(attr_names) < len(attr_values):
            raise DaxClientError(
                'Incorrect number of attributes for attribute list (got {}; expected {})'.format(len(attr_values), len(attr_names)),
                DaxErrorCode.MalformedResult)
        values = dict(zip(attr_names, attr_values))

    item.update(values)

    del item['_anonymous_attribute_values']
    del item['_attr_list_id']
    return item

def _decode_item_internal(dec, request, table_name=None):
    if dec.try_decode_null():
        return None

    t = dec.peek()
    if is_major_type(t, CborTypes.TYPE_MAP):
        proj_ordinals = _get_projection_ordinals(request, table_name)
        item = _decode_projection(dec, proj_ordinals)
    elif is_major_type(t, CborTypes.TYPE_BYTES):
        item = _decode_stream_item(dec.decode_cbor())
    elif is_major_type(t, CborTypes.TYPE_ARRAY):
        item = _decode_scan_result(dec, request.key_schema)
    else:
        raise DaxClientError("Unknown Item type: " + hex(major_type(t)), DaxErrorCode.MalformedResult)

    _reinsert_key(item, request)
    return item

def _get_projection_ordinals(request, table_name):
    proj_ordinals = request.get('_projection_ordinals')
    if proj_ordinals:
        return proj_ordinals

    # Batch requests
    table_proj_ordinals = request.get('_projection_ordinals_by_table')
    if table_proj_ordinals:
        if not table_name:
            raise DaxClientError('Batch request item decoding must include table name', DaxErrorCode.MalformedResult)

        return table_proj_ordinals.get(table_name, [])

    raise DaxClientError('Non-projected request has projected response', DaxErrorCode.MalformedResult)

def _decode_projection(dec, proj_ordinals):
    if proj_ordinals is None:
        raise DaxClientError('Projection ordinals list must not be null', DaxErrorCode.MalformedResult)

    builder = ItemBuilder()
    # Projections are a map[int, AttributeValue] where the key is an index into
    # projection ordinals collected during the request phase
    for _dec in dec.decode_map_iter():
        ordinal = _dec.decode_int()
        try:
            path = proj_ordinals[ordinal]
        except IndexError:
            raise DaxClientError("Unknown projection ordinal: " + ordinal, DaxErrorCode.MalformedResult)

        av = _decode_attribute_value(_dec)
        builder.with_value(path, av)

    return builder.build()

def _decode_stream_item(dec):
    attr_list_id = dec.decode_int()
    anon_attr_values = _decode_anonymous_streamed_values(dec)

    return {
        '_attr_list_id': attr_list_id,
        '_anonymous_attribute_values': anon_attr_values,
    }

def _decode_stream_item_projection(dec):
    # only a partial item is present that will be reconstructed during
    # de-anonymization, when the attrList is available
    # so for now, store the attributes in a map indexed by the ordinal

    attr_list_id = dec.decode_int()
    anon_attr_values = {}
    for _dec in dec.decode_map_iter():
        ordinal = _dec.decode_int()
        anon_attr_values[ordinal] = _decode_attribute_value(_dec)

    # If there are no values (which happens when UPDATED_OLD/NEW have no changes)
    # Pretend that there is no attrListId
    # This will result in an empty Attributes list, which is not what DDB proper does
    # It returns the changed attributes, even if the attributes didn't actually change
    if anon_attr_values:
        return {
            '_attr_list_id': attr_list_id,
            '_anonymous_attribute_values': anon_attr_values,
        }
    else:
        return {}

def _decode_scan_result(dec, key_schema):
    # check for array of length 2
    size = dec._decode_array_header()
    if size != 2:
        raise DaxClientError("Invalid scan item length {} (expected 2)".format(size), DaxErrorCode.MalformedResult)

    item = {}
    # Array item 1 -> key
    key = _decode_key_bytes(dec, key_schema)
    item.update(key)

    # Array item 1 -> value
    value = _decode_scan_value(dec)
    item.update(value)

    return item

def _decode_key_bytes(dec, key_schema):
    key = {}
    hash_attr = key_schema[0]
    hash_attr_type = hash_attr['AttributeType']
    hash_attr_name = hash_attr['AttributeName']
    if len(key_schema) == 1:
        if hash_attr_type == 'S':
            value = dec.decode_binary().decode('utf8')
        elif hash_attr_type == 'N':
            value = str(dec.decode_cbor().decode_number())
        elif hash_attr_type == 'B':
            value = dec.decode_binary()
        else:
            raise DaxClientError("Hash key must be S, B or N, got " + hash_attr_type, DaxErrorCode.MalformedResult)

        key[hash_attr_name] = {hash_attr_type: value}
    elif len(key_schema) == 2:
        key_dec = dec.decode_cbor()
        if hash_attr_type == 'S':
            hash_value = key_dec.decode_string()
        elif hash_attr_type == 'N':
            hash_value = str(key_dec.decode_number())
        elif hash_attr_type == 'B':
            hash_value = key_dec.decode_binary()
        else:
            raise DaxClientError("Hash key must be S, B or N, got " + hash_attr_type, DaxErrorCode.MalformedResult)

        key[hash_attr_name] = {hash_attr_type: hash_value}

        range_bytes = key_dec.drain()

        range_attr = key_schema[1]
        range_attr_type = range_attr['AttributeType']
        range_attr_name = range_attr['AttributeName']

        if range_attr_type == 'S':
            range_value = range_bytes.decode('utf8')
        elif range_attr_type == 'N':
            range_value = str(LexDecimal.decode_all(range_bytes))
        elif range_attr_type == 'B':
            range_value = range_bytes
        else:
            raise DaxClientError("Range key must be S, B or N, got " + range_attr_type, DaxErrorCode.MalformedResult)

        key[range_attr_name] = {range_attr_type: range_value}
    else:
        raise DaxClientError(
            "Key schema must be of length 1 or 2; got {} ({})".format(len(key_schema), key_schema),
            DaxErrorCode.MalformedResult)

    return key

def _decode_compound_key(dec):
    # Compund keys ignore the key schema and simply encode what is given
    # Used for indexed Scan/Query

    key = {}
    for _dec in dec.decode_map_iter():
        name = _dec.decode_string()
        value = _decode_attribute_value(_dec)
        key[name] = value

    return key

def _decode_scan_value(dec):
    return _decode_stream_item(dec.decode_cbor())

def _decode_anonymous_streamed_values(dec):
    # There is no delimiter on the item attributes; the AVs are concatenated
    # and must be read until there is no more data.
    values = []
    while True:
        try:
            av = _decode_attribute_value(dec)
        except CborDecoder.NoMoreData:
            break

        values.append(av)

    return values

def _decode_attribute_value(dec):
    t = dec.peek()

    if is_major_type(t, CborTypes.TYPE_ARRAY):
        return {'L': [_decode_attribute_value(_dec) for _dec in dec.decode_array_iter()]}
    elif is_major_type(t, CborTypes.TYPE_MAP):
        # Dynamo only supports maps of String -> AttributeValue
        av_map = {}
        for _dec in dec.decode_map_iter():
            name = dec.decode_string()
            av = _decode_attribute_value(_dec)
            av_map[name] = av

        return {'M': av_map}
    else:
        v = dec.decode_object()
        if v is None:
            return {'NULL': True}
        if v is True or v is False:
            return {'BOOL': v}
        elif isinstance(v, STRING_TYPES):
            return {'S': v}
        elif isinstance(v, BINARY_TYPES):
            return {'B': v}
        elif isinstance(v, NUMBER_TYPES):
            return {'N': str(v)}
        elif isinstance(v, DdbSet):
            # Set types are wrapped with a wrapper class _Set before being returned
            # This tracks the values and type
            return {v.type: v.values}
        else:
            raise DaxClientError('Unknown type', DaxErrorCode.MalformedResult)

def _reinsert_key(item, request, key_schema=None):
    # Handle GetItem, UpdateItem, DeleteItem
    if 'Key' in request and '_projection_ordinals' not in request:
        # The key attributes are only added if it's NOT a projection
        item.update(request['Key'])
        return

    # Handle PutItem
    if 'Item' in request:
        for key_attr in key_schema:
            key_attr_name = key_attr['AttributeName']
            if key_attr_name not in request['Item']:
                raise DaxClientError(
                        'Request Item is missing key attribute "{}".'.format(key_attr_name),
                        DaxErrorCode.MalformedResult)

            item[key_attr_name] = request['Item'][key_attr_name]

        return



import random
import functools
import time

from .DaxError import DaxErrorCode, DaxClientError, DaxServiceError, is_retryable, is_retryable_with_backoff
from .Constants import SDK_DEFAULT_BASE_DELAY_MS


def equal_jitter_backoff(cap, base=SDK_DEFAULT_BASE_DELAY_MS, attempts=0):
    """Calculate time to sleep based on equal jitter and backoff approach
    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/

    The format is::
        temp = min(cap, base * 2 ** attempt)
        sleep = temp / 2 + random(0, temp / 2)
    If ``base`` is not set then the default value will be used.
    Base must be greater than 0 and a number, otherwise a ValueError will be
    raised.
    """

    if base <= 0:
        raise ValueError("The 'base' param must be greater than 0 and a number, "
                         "got: %s" % base)

    temp = min(cap, base * 2 ** attempts)
    time_to_sleep = temp / 2 + random.randint(0, temp / 2)
    return time_to_sleep


def create_equal_jitter_backoff_function(cap, base):
    """Create an equal jitter function based on the attempts.
    This is used so that you only have to pass it the attempts
    parameter to calculate the delay.
    """

    return functools.partial(
        equal_jitter_backoff, cap, base=base)


class RetryHandler(object):
    DEFAULT_RETRIES = 2  # Same as other DAX clients
    THROTTLED_BASE_DELAY_MS = 70  # Same as other DAX clients
    MAX_BACKOFF_MS = 20 * 1000  # Same as other DAX clients

    def __init__(self, config, cluster):
        self._delay = create_equal_jitter_backoff_function(
            RetryHandler.MAX_BACKOFF_MS, RetryHandler.THROTTLED_BASE_DELAY_MS)
        self._attempts = 0
        self._max_attempts = config.retries.get('max_attempts', RetryHandler.DEFAULT_RETRIES) if config and config.retries \
            else RetryHandler.DEFAULT_RETRIES
        self._retry_timeout = config.connect_timeout
        self._cluster = cluster
        self._last_exception = None

    def can_retry(self, exc):
        """
        Indicates whether the last request can be retried.
        @return true if the last request can be retried, false otherwise
        """
        if self._attempts > self._max_attempts:
            return False

        return is_retryable(exc)

    def pause_before_retry(self, exc):
        self._attempts += 1
        if (not is_retryable_with_backoff(exc)):
            return
        self.wait_for_cluster(exc)
        delayMs = self._delay(attempts=self._attempts)
        return time.sleep(delayMs)

    def wait_for_cluster(self, exc):
        if not isinstance(exc, (DaxClientError, DaxServiceError)):
            return

        if exc.code == DaxErrorCode.NoRoute:
            self._cluster.wait_for_routes(
                min_healthy=1, leader_min=1, timeout=self._retry_timeout)

    def on_exception(self, exc):
        # This function is used to keep in sync with the functionlity of Java Client,
        #   and if the Python client ever has a use case for this.
        self._last_exception = exc



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .CborSExprGenerator import CborSExprGenerator
from .DaxError import DaxValidationError
from botocore.validate import ParamValidator
from six import iteritems
import re

MAX_EXPRESSION_SIZE = 4096
MAX_PARAMETER_MAP_KEY_SIZE = 255
MAX_PARAMETER_MAP_ENTRIES = 2097152
MAX_ATTRIBUTE_NAME_SIZE = 65535


class RequestValidator:
    _PARAM_VALIDATOR = ParamValidator()

    @staticmethod
    def validate_key(item, key_schema):
        if not item:
            raise RequestValidator.new_validation_exception(
                'Value null at %s failed to satisfy constraint: Member must not be null' % item)

        if not len(item.keys()) is len(key_schema):
            raise RequestValidator.new_validation_exception('The number of conditions on the keys is invalid')

    @staticmethod
    def validate_expr_attr_names(attr_names):
        attr_name_map_size = 0

        if len(attr_names.keys()) == 0:
            raise RequestValidator.new_validation_exception('ExpressionAttributeNames must not be empty')
        elif len(attr_names.keys()) >= MAX_PARAMETER_MAP_ENTRIES:
            raise RequestValidator.new_validation_exception('ExpressionAttributeNames exceeds max size')
        else:
            attrs = set()
            for k, v in attr_names.items():
                if k is None:
                    raise RequestValidator.new_validation_exception('ExpressionAttributeNames contains invalid key: null')
                if v in attrs:
                    raise RequestValidator.new_validation_exception('Two document paths overlap with each other; must remove or rewrite '
                                                                    'one of these paths %s' % v)

                attrs.add(v)
                attr_name_map_size += len(k)
                if v is None:
                    attr_name_map_size += 0
                elif len(v) == 0:
                    raise RequestValidator.new_validation_exception('ExpressionAttributeNames contains invalid value: for key %s' % k)
                elif len(v) > MAX_ATTRIBUTE_NAME_SIZE:
                    raise RequestValidator.new_validation_exception('Member must have length less than or equal to %d, Member must have '
                                                                    'length greater than or equal to 0' % MAX_ATTRIBUTE_NAME_SIZE)
                else:
                    attr_name_map_size += len(v)

                if len(k) == 0:
                    raise RequestValidator.new_validation_exception(
                        'ExpressionAttributeNames contains invalid key: The expression attribute map contains an empty key')
                elif k[0] != CborSExprGenerator.ATTRIBUTE_NAME_PREFIX:
                    raise RequestValidator.new_validation_exception('Syntax error, ExpressionAttributeNames contains invalid key: %s' % k)
                elif len(k) > MAX_PARAMETER_MAP_KEY_SIZE:
                    raise RequestValidator.new_validation_exception(
                        'ExpressionAttributeNames contains invalid key: The expression attribute map contains a key that is too long')
                elif not v:
                    raise RequestValidator.new_validation_exception('ExpressionAttributeNames must not be empty')
        return attr_name_map_size

    @staticmethod
    def validate_expr_attr_values(attr_vals):
        attr_values_map_size = 0
        if len(attr_vals.keys()) == 0:
            raise RequestValidator.new_validation_exception('ExpressionAttributeValues must not be empty')
        elif len(attr_vals.keys()) >= MAX_PARAMETER_MAP_ENTRIES:
            raise RequestValidator.new_validation_exception('ExpressionAttributeValues exceeds max size')
        else:
            for k, v in attr_vals.items():
                if k == None:
                    raise RequestValidator.new_validation_exception('ExpressionAttributeValues contains invalid key: null')
                attr_values_map_size += len(k)
                attr_values_map_size += RequestValidator.simple_attr_val_length(v)
                if len(k) == 0:
                    raise RequestValidator.new_validation_exception(
                        'ExpressionAttributeValues contains invalid key: The expression attribute map contains an empty key')
                elif k[0] != CborSExprGenerator.ATTRIBUTE_VALUE_PREFIX:
                    raise RequestValidator.new_validation_exception('Syntax error, ExpressionAttributeValues contains invalid key: %s' % k)
                elif len(k) > MAX_PARAMETER_MAP_KEY_SIZE:
                    raise RequestValidator.new_validation_exception(
                        'ExpressionAttributeValues contains invalid key: The expression attribute map contains a key that is too long')
                RequestValidator.validate_attribute_value(v)
            return attr_values_map_size

    @staticmethod
    def simple_attr_val_length(v):
        if v is None:
            return 0
        if v.get('S'):
            return len(v['S'])
        if v.get('B'):
            return len(v['B'])
        if v.get('N'):
            return len(v['N'])
        if v.get('BS'):
            size = 0
            for b in v['BS']:
                size += len(b)
            return size
            # Only the primitive types are expected
        return 0

    @staticmethod
    def validate_attribute_value(attr):
        # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Limits.html#limits-attributes
        if attr is None:
            return
        if attr.get('S') and len(attr.get('S')) == 0:
            raise RequestValidator.new_validation_exception(
                'One or more parameter values were invalid: An AttributeValue may not contain an empty string')
        elif attr.get('SS'):
            if len(attr['SS']) == 0:
                raise RequestValidator.new_validation_exception('One or more parameter values were invalid: An string set may not be empty')
            for s in attr['SS']:
                if not s:
                    RequestValidator.new_validation_exception(
                        'One or more parameter values were invalid: An string set may not have a empty string as a member')
        elif attr.get('BS'):
            if len(attr['BS']) == 0:
                raise RequestValidator.new_validation_exception(
                    'One or more parameter values were invalid: Binary sets should not be empty')
        elif attr.get('NS'):
            if len(attr['NS']) == 0:
                raise RequestValidator.new_validation_exception(
                    'One or more parameter values were invalid: An number set  may not be empty')
        elif attr.get('M'):
            RequestValidator.validate_item(attr['M'])
        elif attr.get('L'):
            for av in attr['L']:
                RequestValidator.validate_attribute_value(av)

    @staticmethod
    def validate_item(item):
        for k, v in item.items():
            if k is None:
                if v is None:
                    # For some reason DDB SDK accept (null: null) but deny (null: av). (null: null) doesn't represent anything in actual item.
                    raise RequestValidator.new_validation_exception(
                        "Unable to marshall request to JSON: Unable to marshall request to JSON: Unable to marshall request to JSON: Unable to marshall request to JSON")
                else:
                    item.pop(k)
            RequestValidator.validate_attribute_value(v)

    @staticmethod
    def validate_expression(cond_expr=None, upd_expr=None, proj_expr=None, filter_expr=None, key_cond_expr=None, cond_op=None,
                            exp_attr_vals=None, attr_updates=None, attributes_to_get=None, query_filter=None, scan_filter=None,
                            key_condition=None, attr_names=None, attr_vals=None):
        attr_name_map_size = 0
        if attr_names:
            if not cond_expr and not upd_expr and not proj_expr and not filter_expr and not key_cond_expr:
                raise RequestValidator.new_validation_exception('ExpressionAttributeNames can only be specified when using expressions')
            attr_name_map_size = RequestValidator.validate_expr_attr_names(attr_names)
            if attr_name_map_size > MAX_PARAMETER_MAP_ENTRIES:
                raise RequestValidator.new_validation_exception('ExpressionAttributeNames exceeds max size')

        attr_values_map_size = 0
        if attr_vals:
            if not cond_expr and not upd_expr and not filter_expr and not key_cond_expr:
                RequestValidator.new_validation_exception('ExpressionAttributeValues can only be specified when using expressions')
            attr_values_map_size = RequestValidator.validate_expr_attr_values(attr_vals)
            if attr_values_map_size > MAX_PARAMETER_MAP_ENTRIES:
                raise RequestValidator.new_validation_exception('ExpressionAttributeValues exceeds max size')

        if (attr_name_map_size + attr_values_map_size) > MAX_PARAMETER_MAP_ENTRIES:
            raise RequestValidator.new_validation_exception(
                'Combined size of ExpressionAttributeNames and ExpressionAttributeValues exceeds max size')

        if exp_attr_vals is not None:
            for k, v in exp_attr_vals.items():
                if v and v['Value'] and v['AttributeValueList']:
                    raise RequestValidator.new_validation_exception('One or more parameter values were invalid: Value and '
                                                                    'AttributeValueList cannot be used together for Attribute: %s' % k)

        if cond_expr or upd_expr:
            if attr_updates or cond_op or exp_attr_vals:
                raise RequestValidator.new_validation_exception(
                    'Can not use both expression and non-expression parameters in the same request')
            if cond_expr is not None:
                if len(cond_expr) == 0:
                    raise RequestValidator.new_validation_exception('Invalid ConditionExpression: The expression can not be empty')
                elif len(cond_expr) > MAX_EXPRESSION_SIZE:
                    RequestValidator.new_validation_exception(
                        'Invalid ConditionExpression: Expression size has exceeded the maximum allowed size (%d)' % MAX_EXPRESSION_SIZE)
            if upd_expr is not None:
                if len(upd_expr) == 0:
                    RequestValidator.new_validation_exception('Invalid UpdateExpression: The expression can not be empty')
                elif len(upd_expr) > MAX_EXPRESSION_SIZE:
                    RequestValidator.new_validation_exception(
                        'Invalid UpdateExpression: Expression size has exceeded the maximum allowed size (%d)' % MAX_EXPRESSION_SIZE)

        if key_cond_expr is not None:
            if len(key_cond_expr) == 0:
                raise RequestValidator.new_validation_exception('Invalid KeyConditionExpression: The expression can not be empty')
            elif len(key_cond_expr) > MAX_EXPRESSION_SIZE:
                raise RequestValidator.new_validation_exception(
                    'Invalid KeyConditionExpression: Expression size has exceeded the maximum allowed size (%d)' % MAX_EXPRESSION_SIZE)

        if proj_expr is not None:
            if len(proj_expr) == 0:
                raise RequestValidator.new_validation_exception('Invalid ProjectionExpression: The expression can not be empty')
            elif RequestValidator.expression_length(proj_expr, attr_names) > MAX_EXPRESSION_SIZE:
                raise RequestValidator.new_validation_exception(
                    'Invalid ProjectionExpression: Expression size has exceeded the maximum allowed size (%d)' % MAX_EXPRESSION_SIZE)

        if filter_expr is not None:
            if len(filter_expr) == 0:
                raise RequestValidator.new_validation_exception('Invalid FilterExpression: The expression can not be empty')
            elif len(filter_expr) > MAX_EXPRESSION_SIZE:
                raise RequestValidator.new_validation_exception(
                    'Invalid FilterExpression: Expression size has exceeded the maximum allowed size (%d)' % MAX_EXPRESSION_SIZE)

        if proj_expr or filter_expr or key_cond_expr:
            non_expr_params = ''
            expr_params = ''
            # This is required by some of the tests. Probably over-matching but whatever.
            # The order is important. don't re-arrange unless necessary
            if proj_expr:
                expr_params = RequestValidator.append_parameter_name(expr_params, 'ProjectionExpression')
            if filter_expr:
                expr_params = RequestValidator.append_parameter_name(expr_params, 'FilterExpression')
            if key_cond_expr:
                expr_params = RequestValidator.append_parameter_name(expr_params, 'KeyConditionExpression')
            if attributes_to_get:
                non_expr_params = RequestValidator.append_parameter_name(non_expr_params, 'AttributesToGet')
            if scan_filter:
                non_expr_params = RequestValidator.append_parameter_name(non_expr_params, 'ScanFilter')
            if query_filter:
                non_expr_params = RequestValidator.append_parameter_name(non_expr_params, 'QueryFilter')
            if cond_op:
                non_expr_params = RequestValidator.append_parameter_name(non_expr_params, 'ConditionalOperator')
            if key_cond_expr and key_condition:
                non_expr_params = RequestValidator.append_parameter_name(non_expr_params, 'KeyConditions')
            if non_expr_params:
                raise RequestValidator.new_validation_exception(
                    'Can not use both expression and non-expression parameters in the same request: Non-expression parameters: '
                    '{{non_expr_params}} Expression parameters: {{expr_params }}'.format(non_expr_params=non_expr_params,
                                                                                         expr_params=expr_params))

        if cond_op:
            if not exp_attr_vals and not query_filter and not scan_filter:
                raise RequestValidator.new_validation_exception('ConditionalOperator cannot be used without Filter or Expected')
            if (exp_attr_vals and len(exp_attr_vals.keys()) <= 1) or ((query_filter is not None and len(query_filter.keys()) <= 1) or (
                    scan_filter is not None and len(scan_filter.keys()) <= 1)):
                raise RequestValidator.new_validation_exception(
                    'ConditionalOperator can only be used when Filter or Expected has two or more elements')

    @staticmethod
    def expression_length(expr, subs):
        if not expr:
            return 0
        if not subs:
            return len(expr)
        length = len(expr)
        # from is a reserved keyword in python
        for from_, to in iteritems(subs):
            if len(from_) == 0:
                # this should never happen as 'from_' always has prefix '#'.
                # checking the condition to make the code agnostic to other validations.
                return
            reduced = re.sub(from_, '', expr)
            times = (len(expr) - len(reduced)) / len(from_)
            length -= times * len(from_)
            length += times * (len(to) if to else 0)
            expr = reduced
        return length

    @staticmethod
    def append_parameter_name(params, name):
        return params + ', ' + name if params else name

    @staticmethod
    def new_validation_exception(message):  # FIXME match DDB exception.
        return DaxValidationError(message)

    @staticmethod
    def validate_api_using_operation_model(operation_model, request):
        """
        validate using the dyanmodb api operation model. This validate method adds some overhead (~100 microsec). The speed drops with the size of
        request to validate and also if the request is of write nature.

        This validation doesn't validate the correct use of expressions.
        for e.g. 1. The length of expression name has to be less then 65535 which is not validated by DDB service definition.
                 2. TransactItems should have less then 10 items(it validates that there is at least 1 item though)
                 3. In 1 TransactItem, there should only be 1 operation etc.
        DDB service definition validates that the shape of request is correct and required params are met.
        :param operation_model: operation model of the API function to be validated
        :param request: request data from user
        :return: None if it is valid else raise exception
        """
        input_shape = operation_model.input_shape
        if input_shape is not None:
            report = RequestValidator._PARAM_VALIDATOR.validate(request, operation_model.input_shape)
            if report.has_errors():
                raise RequestValidator.new_validation_exception(report.generate_report())



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import threading
import time

from collections import OrderedDict

class SimpleCache(object):
    ''' THIS CLASS MUST BE THREAD SAFE '''
    def __init__(self, size, fetcher):
        self.size = size
        self._fetcher = fetcher
        self._cache = OrderedDict()
        self._lock = threading.RLock()

    def get(self, key, tube):
        with self._lock:
            while True:
                try:
                    return self._get(key)
                except KeyError:
                    if key in self._cache:
                        # The item is actually there, but is a condition
                        # Try again from the top
                        continue

                    try:
                        value = self._fetch(key, tube)
                    except KeyError:
                        continue

                    self._put(key, value)
                    return value

    def _get(self, key):
        # This must be called from within a lock context
        value = self._cache[key]

        # Conditions are used as sentinels to track when an object is in
        # the process of being fetched, so multiple threads do not repeat
        # work.
        if _is_condition(value):
            value.wait(1.0)
            raise KeyError(key)
        else:
            return value

    def _put(self, key, value):
        # This must be called from within a lock context
        self._cache[key] = value
        if len(self._cache) > self.size:
            # Remove the least-recently-added item
            self._cache.popitem(last=False)

    def _fetch(self, key, tube):
        # This must be called from within a lock context
        # Release the lock while fetching
        cond = threading.Condition(self._lock)
        self._cache[key] = cond

        self._lock.release()
        try:
            new_value = self._fetcher(key, tube)
        except:
            self._lock.acquire()
            # Fetch failed
            # If nothing else has updated the key, remove the condition so
            # that another thread can try again
            if self._cache[key] is cond:
                del self._cache[key]
            cond.notify_all()
            raise

        self._lock.acquire()
        try:
            cur_value = self._cache[key]
            if cur_value is cond:
                # Only update if no one else has overwritten
                self._cache[key] = new_value
                return new_value
            else:
                # Updated by somebody else, pretend it doesn't exist
                raise KeyError(key)
        finally:
            cond.notify_all()

class RefreshingCache(SimpleCache):
    def __init__(self, size, fetcher, ttl_millis, clock=time.time):
        super(RefreshingCache, self).__init__(size, fetcher)
        self.ttl_millis = ttl_millis
        self.clock = clock

    def _get(self, key):
        # This must be called from within a lock context
        value, expiry = super(RefreshingCache, self)._get(key)

        now = self.clock()
        if now >= expiry:
            # If it's expired, remove it and pretend it was never there
            del self._cache[key]
            raise KeyError(key)
        else:
            return value

    def _put(self, key, value):
        # This must be called from within a lock context
        now = self.clock()
        expiry = now + self.ttl_millis/1000.0
        return super(RefreshingCache, self)._put(key, (value, expiry))

# On newer versions Condition is a class, on older it's a function returning a
# _Condition, but either way we need to tell whether an object is one.
CONDITION_CLASS = threading.Condition().__class__

def _is_condition(value):
    return isinstance(value, CONDITION_CLASS)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Type encoding sizes measured in bits.
SIZE_8 = 0b00011000
SIZE_16 = 0b00011001
SIZE_32 = 0b00011010
SIZE_64 = 0b00011011
SIZE_ILLEGAL_1 = 0b00011100
SIZE_ILLEGAL_2 = 0b00011101
SIZE_ILLEGAL_3 = 0b00011110
SIZE_STREAM = 0b00011111

# Upper 3 bits of type header defines the major type.
MAJOR_TYPE_MASK = 0b11100000

# Lower 5 bits of type header defines the minor type.
MINOR_TYPE_MASK = 0b00011111

# Positive integer types.
TYPE_POSINT = 0b00000000 # 0..23
TYPE_POSINT_8 = TYPE_POSINT + SIZE_8
TYPE_POSINT_16 = TYPE_POSINT + SIZE_16
TYPE_POSINT_32 = TYPE_POSINT + SIZE_32
TYPE_POSINT_64 = TYPE_POSINT + SIZE_64

# Negative integer types.
TYPE_NEGINT = 0b00100000 # -1..-24
TYPE_NEGINT_8 = TYPE_NEGINT + SIZE_8
TYPE_NEGINT_16 = TYPE_NEGINT + SIZE_16
TYPE_NEGINT_32 = TYPE_NEGINT + SIZE_32
TYPE_NEGINT_64 = TYPE_NEGINT + SIZE_64

# Byte string types.
TYPE_BYTES = 0b01000000 # 0..23 bytes in length
TYPE_BYTES_8 = TYPE_BYTES + SIZE_8
TYPE_BYTES_16 = TYPE_BYTES + SIZE_16
TYPE_BYTES_32 = TYPE_BYTES + SIZE_32
TYPE_BYTES_64 = TYPE_BYTES + SIZE_64
TYPE_BYTES_STREAM = TYPE_BYTES + SIZE_STREAM

# UTF-8 string types.
TYPE_UTF = 0b01100000 # 0..23 bytes in length
TYPE_UTF_8 = TYPE_UTF + SIZE_8
TYPE_UTF_16 = TYPE_UTF + SIZE_16
TYPE_UTF_32 = TYPE_UTF + SIZE_32
TYPE_UTF_64 = TYPE_UTF + SIZE_64
TYPE_UTF_STREAM = TYPE_UTF + SIZE_STREAM

# Array types.
TYPE_ARRAY = 0b10000000 # 0..23 elements
TYPE_ARRAY_8 = TYPE_ARRAY + SIZE_8
TYPE_ARRAY_16 = TYPE_ARRAY + SIZE_16
TYPE_ARRAY_32 = TYPE_ARRAY + SIZE_32
TYPE_ARRAY_64 = TYPE_ARRAY + SIZE_64
TYPE_ARRAY_STREAM = TYPE_ARRAY + SIZE_STREAM

# Map types.
TYPE_MAP = 0b10100000 # 0..23 element pairs
TYPE_MAP_8 = TYPE_MAP + SIZE_8
TYPE_MAP_16 = TYPE_MAP + SIZE_16
TYPE_MAP_32 = TYPE_MAP + SIZE_32
TYPE_MAP_64 = TYPE_MAP + SIZE_64
TYPE_MAP_STREAM = TYPE_MAP + SIZE_STREAM

# Tagged types.
TYPE_TAG = 0b11000000 # for tag type 0..23
TYPE_TAG_8 = TYPE_TAG + SIZE_8
TYPE_TAG_16 = TYPE_TAG + SIZE_16
TYPE_TAG_32 = TYPE_TAG + SIZE_32
TYPE_TAG_64 = TYPE_TAG + SIZE_64

# Simple and special types.
TYPE_SIMPLE = 0b11100000 # not a real type
TYPE_FALSE = TYPE_SIMPLE + 0b00010100
TYPE_TRUE = TYPE_SIMPLE + 0b00010101
TYPE_NULL = TYPE_SIMPLE + 0b00010110
TYPE_UNDEFINED = TYPE_SIMPLE + 0b00010111
TYPE_SIMPLE_8 = TYPE_SIMPLE + SIZE_8 # next byte specifies type 32..255
TYPE_FLOAT_16 = TYPE_SIMPLE + SIZE_16
TYPE_FLOAT_32 = TYPE_SIMPLE + SIZE_32
TYPE_FLOAT_64 = TYPE_SIMPLE + SIZE_64
TYPE_BREAK = TYPE_SIMPLE + SIZE_STREAM

# A few standard tags.
TAG_DATETIME = 0 # string
TAG_TIMESTAMP = 1 # seconds from epoch
TAG_POSBIGINT = 2
TAG_NEGBIGINT = 3
TAG_DECIMAL = 4
TAG_BIGFLOAT = 5

# return type
RET_INT = 0
RET_UTF = 1
RET_BUF = 2
RET_BIGINT = 3
RET_BIGDEC = 4
RET_FLOAT = 5
RET_TAG = 6
RET_BOOL = 7
RET_NULL = 8
RET_UNDEFINED = 9
RET_MAP_HEADER = 10
RET_ARR_HEADER = 11
RET_STREAM_BREAK = 12

def is_major_type(v, t):
    return  (v & MAJOR_TYPE_MASK) == t

def major_type(v):
    return v & MAJOR_TYPE_MASK

def minor_type(v):
    return v & MINOR_TYPE_MASK

def major_type_name(v):
    ''' Attempt to find major type name for the given value. '''
    t = major_type(v)

    type_names = [name for name, value in globals().items() if name.startswith('TYPE') and value == t]
    return '|'.join(type_names) if type_names else 'UNKNOWN'




# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Taken mainly from http://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
# and the DaxJSClient implementation

from __future__ import unicode_literals

import hmac
import hashlib
import datetime
import six
from collections import namedtuple, OrderedDict

Signature = namedtuple('Signature', 'signature string_to_sign token')

_SIGNED_HEADERS = tuple(sorted(('host', 'x-amz-date')))

def generate_signature(creds, endpoint, region, payload, now=None):
    if not now:
        now = datetime.datetime.utcnow()

    headers = _get_headers(endpoint, now, creds)
    return _get_authorization_header('POST', now, headers, payload, creds, region)

def _get_authorization_header(method, now, headers, payload, creds, region):
    canonical_uri = '/'
    canonical_querystring = ''
    credential_scope = datestamp(now) + '/' + region + '/dax/aws4_request'
    canonical_request = '\n'.join((method, 
        canonical_uri,  
        canonical_querystring, 
        _get_canonical_headers(headers), 
        _get_signed_headers(),  
        _SHA256(payload)))

    string_to_sign = '\n'.join(('AWS4-HMAC-SHA256',
        amzdate(now),  
        credential_scope,  
        _SHA256(canonical_request.encode('utf-8'))))

    signing_key = getSignatureKey(creds.secret_key, datestamp(now), region, 'dax')
    signature = _HmacSHA256(signing_key, string_to_sign)
    token = creds.token if hasattr(creds, 'token') and creds.token else None

    return Signature(signature, string_to_sign, token)

def _SHA256(data):
    return hashlib.sha256(data).hexdigest()

def _HmacSHA256(key, data):
    return six.text_type(hmac.new(key, data.encode('utf-8'), hashlib.sha256).hexdigest())

# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

def getSignatureKey(key, dateStamp, regionName, serviceName):
    kDate = sign(('AWS4' + key).encode('utf-8'), dateStamp)
    kRegion = sign(kDate, regionName)
    kService = sign(kRegion, serviceName)
    kSigning = sign(kService, 'aws4_request')
    return kSigning

def _get_canonical_headers(headers):
    return '\n'.join(h + ':' + headers[h] for h in _SIGNED_HEADERS) + '\n'

def _get_signed_headers():
    return ';'.join(_SIGNED_HEADERS)

def _get_headers(host, now, creds):
    headers = OrderedDict()
    headers['host'] = host[8:] if host.startswith('https://') else host
    headers['x-amz-date'] = amzdate(now)
    if hasattr(creds, 'token') and creds.token:
        headers['x-amz-security-token'] = creds.token

    return headers

def amzdate(t):
    return t.strftime('%Y%m%dT%H%M%SZ')

def datestamp(t):
    # Date w/o time, used in credential scope
    return t.strftime('%Y%m%d') 

